#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <conio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "tacor.h"
#include "armafit.h"

double armafit_data[ARMAFIT_N];
double armafit_chi2[ITER2];
double armafit_max;
armafit_param_t param[ARMAFIT_P];
int nd, np, nc, nap, naq;

tacor_vector_t va;
tacor_vector_t vb;
tacor_vector_t vc;

void armafit_param_init(void)
{
    int i;

    for (i = 1; i <= nap; i++) {
        sprintf(PNAME(i - 1), "a[%d]", i);
        PMIN(i - 1) = -1.0;
        PMAX(i - 1) = 1.0;
        PINIT(i - 1) = 0.0;
    }

    for (i = 1; i <= naq; i++) {
        sprintf(PNAME(i + nap - 1), "b[%d]", i);
        PMIN(i + nap - 1) = -1.0;
        PMAX(i + nap - 1) = 1.0;
        PINIT(i + nap - 1) = 0.0;
    }

    np = nap + naq;
}

double armafit_rnd(double min, double max)
{
    double r;

    r = rand();
    r /= RAND_MAX;
    r = min + r * (max - min);
    return (r);
}

void armafit_param_rnd(void)
{
    int i;

    for (i = 0; i < np; i++)
        param[i].current = armafit_rnd(param[i].minimum, param[i].maximum);
}

void armafit_param_limit(void)
{
    int i, j;
    double min, max, delta, a, b, c;

    for (i = 0; i < np; i++) {
        delta = (param[i].maximum - param[i].minimum) / 4.0;
        for (j = 0, b = 0.0, c = 0.0; j < nc; j++) {
            a = 1 / armafit_chi2[j];
            b += param[i].value[j] * a;
            c += a;
        }
        a = b / c;
        param[i].current = a;
        min = a - delta;
        max = a + delta;
        if (min > param[i].minimum)
            param[i].minimum = min;
        if (max < param[i].maximum)
            param[i].maximum = max;
    }
}

void armafit_param_save(void)
{
    int i;

    for (i = 0; i < np; i++)
        param[i].value[nc] = param[i].current;
}

void armafit_param_best(void)
{
    int i;

    for (i = 0; i < np; i++)
        param[i].current = param[i].value[nc];
}

void armafit_load(void)
{
    int i = 0;
    double a;
    FILE *f;

    f = fopen("armafit.inp", "rt");
    while (!feof(f)) {
        fscanf(f, "%lf %lf\n", &a, &armafit_data[i]);
        i++;
    }
    fclose(f);
    nd = i;
}

void armafit_model(void)
{
    int i;

    for (i = 1; i <= va.n; i++)
        va.data[i] = PP(i - 1);
    for (i = 1; i <= vb.n; i++)
        vb.data[i] = PP(i + va.n - 1);
    tacor_init(&va, &vb, &vc);
    tacor_solve();
}

double armafit_chi(void)
{
    int i;
    double c;

    armafit_model();
    for (i = 0, c = 0.0; i <= vc.n; i++)
        c += pow(armafit_data[i] - vc.data[i], 2.0);
    if ((c > 0.0) && (c < armafit_max))
        return c;
    else
        return armafit_max;
}

void armafit_init(void)
{
    int i;

    srand(time(NULL));
    tacor_vector_init(&va, nap);
    tacor_vector_init(&vb, naq);
    tacor_vector_init(&vc, nd - 1);
    va.data[0] = 1.0;
    vb.data[0] = 1.0;
    armafit_param_init();
    nc = 0;
    for (i = 0; i < np; i++) {
        param[i].current = param[i].initial;
        param[i].value[nc] = param[i].initial;
    }
    armafit_chi2[nc] = pow(10.0 * armafit_chi() + 10.0, 2.0);
    armafit_max = armafit_chi2[nc];
}

void armafit_show(void)
{
    int i;

    printf("\n\n");
    for (i = 0; i < np; i++)
        printf("%s = [%g %g] %g\n", param[i].name,
               param[i].minimum, param[i].maximum,
               param[i].current);
    printf("chi^2 = %g\n\n", armafit_chi2[nc]);
}

void armafit_solve(void)
{
    long i, j;
    double c;

    for (i = 0, j = 0; i < ITER0; i++, j++) {
        armafit_param_rnd();
        c = armafit_chi();
        if (c < armafit_chi2[nc]) {
            nc++;
            armafit_chi2[nc] = c;
            armafit_param_save();
            armafit_show();
            if ((nc >= ITER2) || ((nc >= (ITER2 / 2)) && (j >= ITER1))) {
                armafit_param_limit();
                armafit_param_rnd();
                c = armafit_chi();
                armafit_chi2[nc] = c;
                armafit_param_save();
                j = 0;
                nc = 0;
            }
        }
        if (kbhit()) {
            if (getch() == 13)
                armafit_save();
            else if (getch() == 27)
                break;
        }
    }
}

void armafit_save(void)
{
    armafit_param_best();
    armafit_model();
    tacor_vector_save(&va, "armafit1.out");
    tacor_vector_save(&vb, "armafit2.out");
    tacor_vector_save(&vc, "armafit3.out");
}

void main(int argc, char *argv[])
{
    if (argc == 3) {
        nap = atoi(argv[1]);
        naq = atoi(argv[2]);
        if ((nap > 0) && (nap <= 10) && (naq > 0) && (naq <= 10)) {
            armafit_load();
            armafit_init();
            armafit_solve();
            armafit_save();
        }
    }
}
