#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <conio.h>
#include <math.h>
#include <string.h>
#include "tacor.h"

tacor_vector_t *tacor_a;
tacor_vector_t *tacor_b;
tacor_vector_t *tacor_c;
double G[TACOR_PP][TACOR_PP + 1];

int tacor_vector_init(tacor_vector_t * v, int n)
{
    v->n = n;
    v->data = (double *) calloc(v->n + 1, sizeof(double));
    if (v->data == NULL)
        return ERROR;
    return OK;
}

int tacor_vector_done(tacor_vector_t * v)
{
    if (v->data == NULL)
        return ERROR;
    free(v->data);
    v->n = 0;
    return OK;
}

int tacor_vector_save(tacor_vector_t * v, char *fname)
{
    int i;
    FILE *f;

    f = fopen(fname, "wt");
    if (f == NULL)
        return ERROR;
    for (i = 0; i <= v->n; i++) {
        if (fprintf(f, "%d %g\n", i, v->data[i]) < 0) {
            fclose(f);
            return ERROR;
        }
    }
    fclose(f);
    return OK;
}

double tacor_dfunc(int k)
{
    int i;
    double d;

    if (k < 0)
        d = 0.0;
    else if (k == 0)
        d = tacor_b->data[0];
    else if ((k >= 1) && (k <= tacor_b->n)) {
        for (i = 1, d = 0.0; i <= tacor_a->n; i++)
            d += tacor_a->data[i] * tacor_dfunc(k - i);
        d += tacor_b->data[k];
    } else if (k > tacor_b->n)
        d = 0.0;
    return d;
}

void tacor_init(tacor_vector_t * va, tacor_vector_t * vb, tacor_vector_t * vc)
{
    int i, j;

    tacor_a = va;
    tacor_b = vb;
    tacor_c = vc;

    for (i = 0; i <= tacor_c->n; i++)
        tacor_c->data[i] = 0.0;
    for (i = 0; i <= tacor_a->n; i++)
        for (j = 0; j <= (tacor_a->n + 1); j++)
            G[i][j] = 0.0;
}

void tacor_gauss(void)
{
    int i, j, k, p, q;
    int r[TACOR_PP];
    double a, c;

    for (i = 0; i <= tacor_a->n; i++)
        r[i] = i;
    for (j = 0; j < tacor_a->n; j++) {
        p = j;
        q = j;
        for (i = j; i <= tacor_a->n; i++)
            for (k = j; k <= tacor_a->n; k++)
                if (G[i][k] > G[p][q]) {
                    p = i;
                    q = k;
                }
        for (i = 0; i <= (tacor_a->n + 1); i++) {
            c = G[j][i];
            G[j][i] = G[p][i];
            G[p][i] = c;
        }
        for (i = 0; i <= tacor_a->n; i++) {
            c = G[i][j];
            G[i][j] = G[i][q];
            G[i][q] = c;
        }
        k = r[j];
        r[j] = r[q];
        r[q] = k;
        for (i = (j + 1); i <= tacor_a->n; i++) {
            a = G[i][j] / G[j][j];
            for (k = j; k <= (tacor_a->n + 1); k++)
                G[i][k] -= a * G[j][k];
        }
    }
    for (j = tacor_a->n; j >= 0; j--) {
        a = G[j][tacor_a->n + 1];
        for (k = tacor_a->n; k >= (j + 1); k--)
            a -= G[j][k] * G[k][tacor_a->n + 1];
        G[j][tacor_a->n + 1] = a / G[j][j];
    }
    for (i = 0; i <= tacor_a->n; i++)
        G[r[i]][0] = G[i][tacor_a->n + 1];
}

void tacor_solve(void)
{
    int i, k;
    double sa, sb;

    for (k = 0; k <= tacor_a->n; k++) {
        G[k][k] = 1.0;
        for (i = 1; i <= tacor_a->n; i++)
            G[k][abs(k - i)] -= tacor_a->data[i];
        for (i = 0, sb = 0.0; i <= tacor_b->n; i++)
            sb += tacor_b->data[i] * tacor_dfunc(i - k);
        G[k][tacor_a->n + 1] = sb;
    }
    tacor_gauss();
    for (k = 0; k <= tacor_a->n; k++)
        tacor_c->data[k] = G[k][0];
    for (k = (tacor_a->n + 1); k <= tacor_c->n; k++) {
        for (i = 1, sa = 0.0; i <= tacor_a->n; i++)
            sa += tacor_a->data[i] * tacor_c->data[k - i];
        for (i = 0, sb = 0.0; i <= tacor_b->n; i++)
            sb += tacor_b->data[i] * tacor_dfunc(i - k);
        tacor_c->data[k] = sa + sb;
    }
}

void tacor_done(void)
{
}
